-module(buoy_pool).
-include("buoy_internal.hrl").

-export([
    init/0,
    lookup/3,
    start/1,
    start/2,
    stop/1,
    terminate/0
]).

%% public
-spec init() ->
    ok.

init() ->
    foil:new(?MODULE),
    foil:load(?MODULE).

-spec lookup(protocol_http(), hostname(), inet:port_number()) ->
    {ok, atom()} | {error, pool_not_started | buoy_not_started}.

lookup(Protocol, Hostname, Port) ->
    case foil:lookup(buoy_pool, {Protocol, Hostname, Port}) of
        {ok, _} = R ->
            R;
        {error, key_not_found} ->
            {error, pool_not_started};
        {error, _} ->
            {error, buoy_not_started}
    end.

-spec start(buoy_url()) ->
    ok | {error, pool_already_started | buoy_not_started}.

start(Url) ->
    start(Url, ?DEFAULT_POOL_OPTIONS).

-spec start(buoy_url(), options()) ->
    ok | {error, pool_already_started | buoy_not_started}.

start(#buoy_url {
        protocol = Protocol,
        hostname = Hostname,
        port = Port
    }, Options) ->

    Name = name(Protocol, Hostname, Port),
    ClientOptions = client_options(Protocol, Hostname, Port, Options),
    PoolOptions = pool_options(Options),

    case agAgencyPoolMgr:start(Name, ?CLIENT, ClientOptions, PoolOptions) of
        ok ->
            Key = {Protocol, Hostname, Port},
            ok = foil:insert(?MODULE, Key, Name),
            ok = foil:load(?MODULE);
        {error, pool_already_started} = E ->
            E;
        {error, shackle_not_started} ->
            {error, buoy_not_started}
    end.

-spec stop(buoy_url()) ->
    ok | {error,  pool_not_started | buoy_not_started}.

stop(#buoy_url {
        protocol = Protocol,
        hostname = Hostname,
        port = Port
    }) ->

    Key = {Protocol, Hostname, Port},
    case foil:lookup(?MODULE, Key) of
        {ok, Name} ->
            agAgencyPoolMgr:stop(Name),
            foil:delete(?MODULE, Key),
            foil:load(?MODULE);
        {error, key_not_found} ->
            {error, pool_not_started};
        {error, _} ->
            {error, buoy_not_started}
    end.

-spec terminate() ->
    ok.

terminate() ->
    foil:delete(?MODULE).

%% private
client_options(Protocol, Hostname, Port, Options) ->
    Reconnect = ?LOOKUP(reconnect, Options, ?DEFAULT_IS_RECONNECT),
    ReconnectTimeMax = ?LOOKUP(reconnect_time_max, Options,
        ?DEFAULT_RECONNECT_MAX),
    ReconnectTimeMin = ?LOOKUP(reconnect_time_min, Options,
        ?DEFAULT_RECONNECT_MIN),

    [{ip, binary_to_list(Hostname)},
     {port, Port},
     {protocol, shackle_protocol(Protocol)},
     {reconnect, Reconnect},
     {reconnect_time_max, ReconnectTimeMax},
     {reconnect_time_min, ReconnectTimeMin},
     {socket_options, [
         binary,
         {packet, line},
         {packet, raw},
         {send_timeout, 50},
         {send_timeout_close, true}
     ]}].

name(Protocol, Hostname, Port) ->
    list_to_atom(atom_to_list(Protocol) ++ "_"
        ++ binary_to_list(Hostname) ++ "_"
        ++ integer_to_list(Port)).

pool_options(Options) ->
    BacklogSize = ?LOOKUP(backlog_size, Options, ?DEFAULT_BACKLOG_SIZE),
    PoolSize = ?LOOKUP(pool_size, Options, ?DEFAULT_POOL_SIZE),
    PoolStrategy = ?LOOKUP(pool_strategy, Options, ?DEFAULT_POOL_STRATEGY),

    [{backlog_size, BacklogSize},
     {pool_size, PoolSize},
     {pool_strategy, PoolStrategy}].

shackle_protocol(http) ->
    shackle_tcp;
shackle_protocol(https) ->
    shackle_ssl.
